"""
This is a setup.py script generated by py2applet

Usage:
    python setup.py py2app
"""

from setuptools import setup

APP = ['main.py']
DATA_FILES = []
OPTIONS = {
    'excludes': ['numpy', 'email'],
    'argv_emulation': False,
    'iconfile': 'resources/appIcon.icns',
    'semi_standalone': 'False',
    'optimize': 2,
    'resources': ['client.py', 'sync.py', 'utils_kivy.py', 'utils.py', 'cacert.pem', 'img', 'management', 'aldryngui.kv'],
    'packages': ['certifi', 'wxwidgets', 'wxpython'],
    'site_packages': 'True',
    'plist': {
        'LSPrefersPPC': False,
        'LSArchitecturePriority': 'x86_64',
    }
}

setup(
    name='AldrynCloud',
    version='0.0.1',
    author='Divio AG',
    author_email='info@divio.ch',
    description='App to control static files of Aldryn Cloud sites',
    app=APP,
    data_files=DATA_FILES,
    options={'py2app': OPTIONS},
    setup_requires=['py2app'],
)
